(***************************************************)
(*    HTTcc Examples for the ICFP'13 submbission   *)
(*     Using the finalisation code as a hanlder    *)
(***************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq div.
From Kont Require Import rels prelude heaps domains hprop callcc srules.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ccscope.

Parameter (X : lattice).

(* The following examples aim at illustrating the power of algebraic callcc   *)
(* by showing applications of the use of finalization code with abort.        *)


(* Continuations as handler:  A typical programming idiom with continuations  *)
(* is to use continuations to implement exception handling. The Haskell Wiki  *)
(* presents the following example:                                            *)

(* Example: An exception-throwing div                                         *)

(*  divExcpt :: Int -> Int -> (String -> Cont r Int) -> Cont r Int            *)
(*  divExcpt x y handler =                                                    *)
(*    callCC $ \ok -> do *)
(*      err <- callCC $ \notOk -> do                                          *)
(*        when (y == 0) $ notOk "Denominator 0"                               *)
(*        ok $ x `div` y                                                      *)
(*      handler err                                                           *)
 
(*  {- For example,                                                           *)
(*  runCont (divExcpt 10 2 error) id  -->  5                                  *)
(*  runCont (divExcpt 10 0 error) id  -->  *** Exception: Denominator 0       *)
(*  -}                                                                        *)


(* The idiom consists of using two nested callcc's where one continuation     *)
(* object will be used to pass correct values to the continuation, whilst the *)
(* other will be used to pass an error to a handler.                          *)

(* Using algebraic callcc we can do the same with just one callcc.            *)
(* Silly exception handling using a algebraic callcc *)

Section SimpleHandling.

Inductive Maybe (A: Type) : Type :=
  | nothing : Maybe A
  | just : A -> Maybe A.

Arguments just {A}.

Program
Definition handler : 
  SK X (xPredT, fun y i m => i = m /\ y = @nothing nat) :=
      Do (ret (@nothing nat)).
Next Obligation. by move=> i _ ; apply: val_ret. Qed.


Program
Definition unsafe_div (x y : nat) : 
    SK X (fun i => y <> 0, fun z  i m => i = m /\ z = x %/ y ) :=
    Do (ret (  x %/ y)). 
Next Obligation. by move=> i _ ; apply: val_ret. Qed.


Program 
Definition cont_div (p q: nat) : SK _ 
  (xPredT, fun (y  : Maybe nat) i m => i = m 
            /\ if (q == 0) then  y = (@nothing nat) else y = just (p %/ q)) :=
Do ( c <-- callcc xPred1 
            (fun j y i m => i = m /\ 
                if (q == 0) then  y = (@nothing nat) else y = just (p %/ q))
             (fun j exit => If (q == 0) Then abort _ exit (handler) Else
                             z <-- unsafe_div p q; 
                             ret (just z));
     ret c).
Next Obligation.
move=> i _; apply: bnd_ccc=>//=.
+ case: eqP=>Q; first by apply: bnd_abort=>/=; do 2 split=>//.
 by apply: bnd_bnd; apply:bnd_do=>//= x k [->]->; apply:bnd_ret;apply:val_ret.
by move=> x k [h][_][<-][<-]; case:eqP=>//= _ ->; apply: val_ret.
Qed.
End SimpleHandling. 

(* An old silly example: Swapping and conditionally swapping back in
the finalisation code: *)

Section Swapback.

Program 
Definition swapback n : SK X (xPredT, fun y i m => 
                      i = m /\  if (n >= 2) then y = 100 else y = 0) :=
  Do ( x <-- alloc 0 ;
       y <-- alloc 100 ;
       c <-- callcc (fun i m =>
                              forall h (v w: nat),
                                i = y :-> v :+ x :-> w :+ h ->
                               if (n >= 2) then m = y :-> w :+ x :-> v :+ h
                                else False)
                      (fun j z i m => 
                              forall h (v w: nat),
                                i = y :-> v :+ x :-> w :+ h ->
                                m = y :-> w :+ x :-> v :+ h /\ z = tt)
                     (fun j exit =>
                                a <-- @CCC.read X nat x ;
                                b <-- @CCC.read X nat y ;
                                x ::= b ;;
                                y ::= a ;;
                                If (n >= 2) Then
                                abort _ exit (y ::= b ;; x ::= a)
                                Else
                                ret tt);
       b <-- ! y ;
       dealloc x ;;
       dealloc y ;;
       ret b ).
Next Obligation.
move => i _; apply: bnd_alloc=>x; apply: bnd_alloc=>y;apply: vrfD=>D0.
apply: bnd_ccc=>//=.
+ do 2 (apply: bnd_bnd; rewrite unCA; apply: bnd_read).
 do 2 (apply: bnd_bnd; rewrite unCA; apply: bnd_write); apply: vrfD=>D1.
 case: ifP=>/=H.
 - move:H=>->;apply: bnd_abort; split.
  + split=>//h v w; rewrite unCA -!unA; move/(cancel D0)=>[<-]D2.
    by move/(cancel D2)=>[<- _ ->].
  move=>/= j[[H0 _]]->. 
  apply: bnd_write; rewrite unCA; apply: val_write=> _ h v w.
  rewrite -!unA; move/(cancel D1)=>[<-]D3; move/(cancel D3)=>[<-]_->.
  by rewrite unCA.
 apply: bnd_ret; apply: bnd_read; rewrite unCA; do 2 apply: bnd_dealloc.
 by apply:val_ret.
case=> k; rewrite unA=>[[h]][_];case.
move/(_ i 100 0 (erefl _)); case: ifP=>//H1->; case/(_ i 0 100 (erefl _))=>-> _.
rewrite -unA; apply: bnd_read; rewrite unCA; do 2 apply: bnd_dealloc.
by apply:val_ret.
Qed.
End Swapback.