(***************************************************)
(*    HTTcc Examples for the ICFP'13 submbission   *)
(*               Ping-Pong Cooperation             *)
(***************************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq div.
From Kont Require Import rels prelude heaps domains hprop callcc srules.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ccscope.

Parameter (X : lattice).
(* Ping / Pong Coroutines , Section 6.2. *)

(* We unfold the defininiton of separation star (#), this, \In, etc
and deal with explicit heap equality and heap union :+ :

 p1 # p2 : Pred heap :=
Pred_of_Simpl
  [Pred h | exists h1 h2 : heap, h = h1 :+ h2 /\ h1 \In p1 /\ h2 \In p2]

this h  : Pred heap := Pred_of_Simpl [Pred h | i = h]

i \In P <-> P i

Other relevant definitions and properties can be found in the heaps.v
and hprop.v. For further details on def predicates, see

[26] NANEVSKI, A., VAFEIADIS, V., AND BERDINE, J. Structuring the
verification of heap-manipulating programs. In POPL (2010). *)


(* We implement Ping and Pong trivially by using functions that *)
(* increment  x and y variable by one. *)

(* incr x : [v h] . SK* {x :-> v # this h} () {x :-> v + 1 # this h} *)

Program Definition incr ( x : ptr) : 
   SK X ( fun i => exists h, exists v: nat , i = x :-> v :+ h , 
          fun y i m => forall h v, i = x :-> v :+ h -> 
                         m = x :-> v .+1 :+ h /\ y = tt) :=
Do ( z <-- ! x ; x ::= z .+1 ).
Next Obligation.
move=> i [h][v]->; apply: bnd_read; apply: vrfD=>D; apply:val_write=> _ h1 v1.
by move/(cancel D)=>[->]_->; split=>//.
Qed.


Section PingPong.
Variables x y : ptr.
Definition ping := incr x.
Definition pong := incr y.

(* The ()-spec in the conseq. in PingPongT:  

  s1 = ( x :-> v + p + 1 # y :-> w + p # this h,  
    if  p == n then  
        x :-> v + n + 1 # y :-> w + n +1 # this h 
    else \bot )*

 is translated to (ppT,pptQ) where *)

Definition pptP (v w p : nat) h  : pre :=
    fun i =>  
          i =  x :-> (v + p.+1) :+  y :-> (w + p) :+ h.

Definition pptQ (v w p n: nat) h : post unit := 
  fun r (i m : heap) =>    
      i = x :-> (v + p.+1) :+ y :-> (w + p ) :+ h -> 
      if  p == n then 
        m = x :-> (v + n.+1) :+ y :-> (w + n .+1) :+  h
      else False.

(* PingPong T := forall z.
      [v w p h]. SK* { x :-> v + p # y :-> w + p # this h
􏰂                    _/\_ j \In x:-> v # y :-> w # this h
                      _/\_ z <=  _/\_􏰀 p + z  = n} (SigSK ())
                     { x :-> v + p # y :-> w + p # this h
                      _/\_ spec r \lleq s1 } *)

Definition pingPongT (n : nat) (j :  heap) : Type := 
  forall z: nat , SK X 
     ( fun i => exists h, exists v : nat, exists w: nat, exists p: nat,
                 i = x :-> (v + p) :+  y :-> (w  + p)  :+  h 
                 /\  j  = x :-> v :+  y :-> w :+ h
                 /\ z <= n
                 /\ p + z = n,
       fun (r  : SigSK X) i m  => 
        forall h (v w p: nat),
                 i = x :-> (v + p) :+  y :-> (w  + p)  :+  h ->
                 j  = x :-> v :+  y :-> w :+ h ->
                 m =  x :-> (v + p) :+  y :-> (w  + p) :+ h
                 /\ CCC.conseq (spec_of r) (pptP v w p h) (pptQ v w p n h)).

(* We rename the continuation object k  in Figure 8 to ex1:

ex1 : [v w h]<p>. Kont { j \In x j \In x:-> v # y :-> w # this h}
                       { if n <> 0 then 
                           x :-> v + p + 1 # y :-> v + p # this h
                         else \bot} (SigSK ())
                       {  x :-> v + p + 1 # y :-> v + p +1 # this h
                         _/\_ conseq r \lleq s2 }

where

 s2 = ( x :-> v + p + 2 # y :-> v + p +1 # this h,
       if p + 1 = n then 
          x :-> v + n + 1 # y :-> v + n +1 # this h
       else \bot)*

We translate the spec s2 as  *)

Definition kontP v w p h : pre :=
    fun i => i =  x :-> (v + p.+2) :+  y :-> (w + p.+1) :+ h. 

Definition kontQ (v w p n: nat) h : post unit := 
  fun _ (i m : heap) =>
      i = x :-> (v + p.+2) :+ y :-> (w + p.+1 ) :+ h -> 
      if  p.+1 == n then 
        m = x :-> (v + n.+1) :+ y :-> (w + n .+1) :+ h
      else False.

Program
Definition pingPong ( n : nat) : SK _ ( fun i =>  
    exists v: nat, exists w: nat, exists h, 
              i = x :-> v :+ y :-> w :+ h,
         fun z i m =>  forall (v w: nat) h,
           i = x :-> v  :+ y :-> w :+ h ->
              m = x :-> ( v + n.+1) :+ y :-> (w + n.+1) :+ h):=
  Do ( c <-- callcc 
(* R *)
    (fun j i => forall (v w : nat) h, 
                     j = x :-> v :+ y :-> w :+ h -> 
                     if (n == 0) then False else
                     exists p :nat, 
                     i = x :-> (v + p.+1) :+ y :->  (w + p) :+ h)
(* Q *)
      (fun j z i m => 
                    forall (v w p: nat) h, 
                      j = x :-> v :+ y :-> w :+ h -> 
                      if (n == 0) then False else
                        i = x :-> (v + p.+1) :+ y :-> (w + p) :+  h ->
                      m = x :-> (v + p.+1) :+ y :-> (w + p.+1) :+ h
                      /\
                          CCC.conseq (spec_of z)
                                     (kontP v w p h) (kontQ v w p n h))
 
        (fun j ex1 =>
                 ffix (fun (f: pingPongT n j) (m: nat) => 
                   Do (CCC.Match_nat m 
                       (ret (BigSK pong))
                       (fun z => ret 
                          (BigSK (abort _ ex1 (pong ;; (f z))))))) n);
      ping ;;
      comm c).
Next Obligation.
rewrite !gh; apply:ghE=>/=.
- by move=> i [h][v][w][p][H0][H1] _ _; exists (h,v,w,p).
move=> i s [H0 H1] [h][v][w][p][I][J]HH. 
rewrite I -!unA in H0 * =>D; rewrite J -!unA in H1. 
have D1: def ( x :-> v :+ (y :-> w :+ h))
             by rewrite !defPtUn !domPtUn !defPtUn in D *.
case: (cancel D1 H1) H0=> <- D2; case/(cancel D2)=><- _ <-.
case/(cancel D); move/addnI=> <- _ _ {D D2 s H1 I}.
(* We case on m (we don't need the IH *)
(* m = 0 , thus we are in the  (ret (BigSK (pong))) branch *)
case: m HH=>[[_]|m0].
+ rewrite addn0=> ->; rewrite unA; apply: val_ret=>//=.
 split=>// k ->. 
 (* pong! *) 
 apply: val_do=>//= { k}.  
 - by exists  (x :-> (v + n.+1) :+ h); exists (w + n); rewrite unCA unA.
 case => k {ex1 f}  _.
 move/(_ (x :-> (v + n.+1) :+ h) (w + n)); rewrite unCA unA.
 by case/(_ (erefl _))=> -> _ _ ;case:eqP=>//= _; rewrite  -addnS unCA unA.
(* m <> 0 ) *) 
rewrite addnS=>[[H0]] HH; apply: val_ret; split; first by rewrite unA; hhauto.
(* the body just returns, but we need to prove the conseq for the returnee. *)
(* This is the first layer of Big, namely abort *)
move=> k ->. 
(* we've gone somehere and now abort gets evaluated *)
apply: val_abort; split=>/= {k}.
+ split=>//  v1 w1 h1; rewrite J -!unA; case/(cancel D1)=><- D3.
 case/(cancel D3)=><- _ <- {D3}.
 case: eqP HH=>//= [->|]; first (by discriminate); move=>_ HH;
 by exists p; rewrite -unA unCA unA -unAC -unA unCA unA.
(* What follows is the conseq in abort *)
rewrite !gh; apply: ghE.
+ move=> k [[_] _] -> _ ; by exists (v,w,p,h).
move=> {f ex1} k s; rewrite J -!unA; case/(cancel D1)=><- D3. 
case/(cancel D3)=><- _ <- [[_] _] -> {k D3} D3.
(* pong *)
apply: bnd_do=>{D1}//=.
- by exists  (x :-> (v + p.+1) :+ h); exists (w + p); rewrite unCA.
case=> k; move/(_ (x :-> (v + p.+1) :+ h) (w + p)); rewrite unCA.
case/(_ (erefl _)); rewrite  unCA=>-> _.
(* recursive call , I now there are more jumps, just need to get there*)
apply: val_do=>//=. 
+ exists h; exists v; exists w; exists p.+1.
 by rewrite addnS addnS ltnW //=; rewrite unA; hhauto.
move=> c {k} k _; rewrite -addnS !unA.
case/(_ h v w p.+1 (erefl _) (erefl _)); rewrite -unA=> -> H1. 
case: eqP HH=>//=; first by move=> ->; discriminate.
move=> H2 HH; rewrite unAC unC -!unA; case/(cancel D3)=> _ D4.
by case/(cancel D4); move/addnI=><- _ _ {D3 D4}; split=>//.
Qed.

Next Obligation.
rewrite !gh; apply: ghE; first by move => i [v][w][h] -> _; exists (v,w,h).
move=> i s H0 [v][w][h]; rewrite -!unA in H0 *=>I; rewrite I in H0 *=>D.
case: (cancel D H0)=><- DD; case/(cancel DD)=><- _ <- {s H0 DD}.
apply: bnd_ccc=>/=.
(* Qe exit i.e. first time out *)
+ have P1: pptP v w 0 h (x :-> v.+1 :+  (y :-> w :+  h)).
 -by unfold pptP; rewrite  addn0 addn1 unA.  
 apply: bnd_do=>//=. 
 -  by exists h; exists v; exists w; exists 0; rewrite unA add0n !addn0.
 move=> c k ;move/(_ h v w 0);rewrite !addn0 unA. 
 case/(_ (erefl _) (erefl _))=>->.
  case/(_ (x :-> v.+1 :+ (y :-> w :+ h)) P1)=>[| H2 H3] ;
   first by rewrite !defPtUn !domPtUn !defPtUn in D *.
 (* First ping *)
 apply: bnd_do=>//= {k} [| tt k]. 
 - by exists ( y :-> w :+ h); exists  v; rewrite unA.
 move/(_ (y :-> w :+ h) v); rewrite unA; case/(_ (erefl _))=>-> _.
 apply: val_do=>// {k tt} tt k Dk; move/(H3 _ _ ); move/(_ Dk).
 unfold pptQ; rewrite addn1 addn0 -unA; move/(_ (erefl _)).
 by case: eqP=>//= <- -> ; rewrite unA.
(* R ; qf exit *)
move=>/= c m [k][_]; rewrite unA; case; move/(_ v w h (erefl _)).
(* If n = 0 there was no loop at all *)
case: eqP=>//= H [p] ->.
(* Last time out of the loop *)
have P1: kontP v w p h
              (x :-> (v + p.+2) :+ ( y :-> ( w + p.+1) :+ h)).
+ by unfold kontP; rewrite unA.
case/(_ v w p h (erefl _) (erefl _))=>-> H2;rewrite -unA; apply: vrfD=> D1.
 (* (n + 1) Ping *)
apply: bnd_do=>//= {k} [| tt k]; first by hhauto.
move/(_ (y :-> (w + p.+1) :+ h) (v + p.+1) (erefl _))=>[->]_. 
apply: vrfD;rewrite -addnS.
case/(H2 (x :-> (v + p.+2) :+ ( y :-> (w + p.+1) :+ h)) P1)=> P2 H3.
apply: val_do=>// {tt k} tt k Dk; move/(H3 _ _); move/(_ Dk).
unfold kontQ; rewrite unA; move/(_ (erefl _)).
(* If I'm not at the last iteration, niceQ will be False. Thus, the only jump *)
(* I'm interested in is the last time out. *)
by case:eqP =>// <- ->; rewrite unA.
Qed.
End PingPong.

